#include "PyEMData.h"
#include "PyPickle.h"
#include <boost/python/detail/extension_class.hpp>
#include <boost/python/reference.hpp>
#include <stdio.h>

#define PY_ARRAY_UNIQUE_SYMBOL Py_Array_pyEM_EMData
#define NO_IMPORT_ARRAY
#include <Numeric/arrayobject.h>

#ifdef DEBUG
#include "assert.h"
#endif

BOOST_PYTHON_BEGIN_CONVERSION_NAMESPACE
static python_extension_class_converters<PyEMData> _PyEMData_type;
static python_extension_class_converters<PyEuler> _PyEuler_type;
BOOST_PYTHON_END_CONVERSION_NAMESPACE

void pylist2emlist(const PyList& pylist, List* emlist)
{
    if (!emlist) {
	return;
    }
    
#ifdef DEBUG
    assert(emlist->count() == 0);
#endif
    
    for (int i = 0; i < pylist.size(); i++) {
	PyObject* obj = pylist.get_item(i).get();
	PyEMData* item = BOOST_PYTHON_CONVERSION::from_python(obj, boost::python::type<PyEMData*>());
	emlist->addObject(item);
    }
    
}

void emlist2pylist(List* emlist, PyList& pylist)
{    
    if (!emlist) {
	return;
    }
    
#ifdef DEBUG
    assert(pylist.size() == 0);
#endif
    
    for (int i = 0; i < emlist->count(); i++) {
	PyEMData item = *(static_cast<PyEMData*>(emlist->objectAt(i)));
	PyObject* obj = BOOST_PYTHON_CONVERSION::to_python(item);
	pylist.append(boost::python::ref(obj));
    }
}



int fileCount(const char* filespec, String& filetype)
{
    int type = 0;
    int nimg = fileCount(filespec, &type);
    filetype.val = "unknown";

    if (type == 0) {
        filetype.val = "mrc/other";
    }
    else if (type == 1) {
        filetype.val = "spider";
    }
    else if (type == 2) {
        filetype.val = "imagic";
    }
    else if (type == 4) {
        filetype.val = "pgm/gif";
    }
    else if (type == 5) {
        filetype.val = "lst";
    }
    else if (type == 6) {
        filetype.val = "pif";
    }

    return nimg;
}

static void get_dims(int nx, int ny, int nz, int* p_ndims, int** p_dims)
{
    (*p_ndims) = 0;
    
    if (nx > 1) {
	(*p_ndims)++;
    }
    if (ny > 1) {
	(*p_ndims)++;
    }
    if (nz > 1) {
	(*p_ndims)++;
    }

    int* dims = new int[*p_ndims];
    int i = 0;

    if (nx > 1) {
	dims[i] = nx;
	i++;
    }
    if (ny > 1) {
	dims[i] = ny;
	i++;
    }
    if (nz > 1) {
	dims[i] = nz;
    }
    *p_dims = dims;    
}


PyEMData PyEMData::null_emdata = PyEMData();
Float PyEMData::null_float = Float();
XYData PyEMData::null_xydata = XYData();


PyEMData::PyEMData() : EMData()
{
    
}


PyEMData::~PyEMData()
{
    
}

PyEMData* PyEMData::deref()
{
    if ((int) this == (int) &null_emdata) {
	return 0;
    }
    return this;
}


float* PyEMData::deref(Float& f)
{
    if ((int)&f == (int)&null_float) {
	return 0;
    }
    return &f.val;
}

XYData* PyEMData::deref(XYData& xy)
{
    if ((int)&xy == (int)&null_xydata) {
	return 0;
    }
    return &xy;
}


PyObject* PyEMData::get_num_array()
{
    int ndim;
    int* dims;
    
    get_dims(xSize(), ySize(), zSize(), &ndim, &dims);

    float* array = getData();
    
    PyArrayObject* num_array =
	(PyArrayObject*) PyArray_FromDimsAndData(ndim, dims,
						 PyArray_FLOAT, (char*)array);

#if 0
    float* num_array_data = (float*) num_array->data;
    int sz = xSize() * ySize() * zSize();

    assert(array != 0);
        
    for (int i = 0; i < sz; i++) {
	num_array_data[i] = array[i];
    }
    
#endif
    
    doneData();
    delete [] dims; dims = 0;
    return (PyObject*) num_array;
}

void PyEMData::set_num_array(PyObject* num_array, int array_sz)
{
#ifdef DEBUG
    assert(num_array != 0);
#endif
    
    float* num_array_data = (float*)((PyArrayObject*)num_array)->data;
    float* array = getData();

#ifdef DEBUG
    assert(array != 0);
#endif
    
    for (int i = 0; i < array_sz; i++) {
	array[i] = num_array_data[i];
    }
    
    doneData();
}


// return 0 if OK; 
int PyEMData::readImage(const char* filename, ImageType img_type, int n, int nodata)
{
    if (!filename) {
	printf("Error: no file is input\n");
	return -1;
    }

    switch(img_type) {
    case IMAGE_ANY:
	return EMData::readImage(filename, n, nodata);
    case IMAGE_SAL:
	return readSAL(filename, nodata);
    case IMAGE_PGM:
	return readPGM(filename, nodata);
    case IMAGE_ICOS:
	return readICOS(filename, nodata);
    case IMAGE_MRC:
	return readMRC(filename, nodata);
    case IMAGE_IMAGIC:
	return readIMAGIC(filename, n, nodata);
    case IMAGE_IMAGIC3D:
	return readIMAGIC3d(filename, nodata);
    case IMAGE_EMIM:
	return readEMIM(filename, n, nodata);
    case IMAGE_SPIDER:
	return readSPIDER(filename, n, nodata);
    case IMAGE_GATAN:
	return readGatan(filename, nodata);
    case IMAGE_GATAN3:
	return readGatan3(filename, nodata);
    case IMAGE_PIF:
	return readPIF(filename, n, nodata);
    default:
	printf("Error: invalid file type '%d' for file '%s'\n",
	       img_type, filename);
	return -1;
    }
    
    return 0;
}

// return 0 if OK; 
int PyEMData::writeImage(const char* filename, ImageType img_type,
			 int n, int nolock, int mode)
{
    if (!filename) {
	printf("Error: no file is input\n");
	return -1;
    }

    switch(img_type) {
    case IMAGE_ANY:
	printf("Error: please specify image type\n");
	return -1;    
    case IMAGE_IMAGIC3D:
	return writeIMAGIC3D(filename, nolock);
    case IMAGE_IMAGIC:
	return writeIMAGIC(filename, n, nolock);
    case IMAGE_ICOS:
	return writeICOS(filename);
    case IMAGE_MRC:
	return writeMRC(filename, mode);
    case IMAGE_SPIDER:
	return writeSPIDER(filename, n);
    case IMAGE_SINGLE_SPIDER:
	return writeSingleSPIDER(filename);
    case IMAGE_EMIM:
	return writeEMIM(filename, n);
    case IMAGE_VTK:
	return writeVTK(filename);
    case IMAGE_AMIRA:
	return writeAMIRA(filename);
    default:
	printf("Error: invalid file type '%d' for file '%s'\n",
	       img_type, filename);
	return -1;
    }
    
    return 0;
}


PyEMData_ptr PyEMData::py_copy(int withfft, int withparent)
{
    PyEMData* em = static_cast<PyEMData*> (copy(withfft, withparent));
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_copyHead()
{
    PyEMData* em = static_cast<PyEMData*> (copyHead());
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_clip(int x0, int y0, int w, int h)
{
    PyEMData* em = static_cast<PyEMData*> (clip(x0, y0, w, h));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_clip(int x0, int y0, int z0, int w, int h, int d)
{
    PyEMData* em = static_cast<PyEMData*> (clip(x0, y0, z0, w, h, d));
    return PyEMData_ptr(em);
}

void PyEMData::py_insertClip(PyEMData& block, int x0,  int y0)
{
    insertClip(block.deref(), x0, y0);
}

    
void PyEMData::py_calcHist(PyList hist_list, int n, float hmin,
			   float hmax, int add)
{
#ifdef DEBUG
    assert(hist_list.size() >= n);
#endif
    
    float* hist_array = new float[hist_list.size()];
    list2array<float>(hist_list, hist_array);

    calcHist(hist_array, n, hmin, hmax, add);

    array2list<float>(hist_array, hist_list);
    delete [] hist_array;
    hist_array = 0;
}


void PyEMData::py_setParent(PyEMData& data)
{
    setParent(data.deref());
}

PyEMData_ptr PyEMData::py_Parent()
{
    PyEMData* em = static_cast<PyEMData*> (Parent());
    return PyEMData_ptr(em);
}

PyEuler PyEMData::py_getEuler()
{
    PyEuler* e = static_cast<PyEuler*> (getEuler());
    return *e;
}

void PyEMData::py_setRAlign(PyEuler angle)
{
    setRAlign(&angle);
}


imagicH_ptr PyEMData::py_getImagicHed()
{
    imagicH* ih = getImagicHed();
    return imagicH_ptr(ih);
}

void PyEMData::py_setImagicHed(imagicH& hed)
{
    setImagicHed(&hed);
}

mrcH_ptr PyEMData::py_getMRCHed()
{
    mrcH* mh = getMRCHed();
    return mrcH_ptr(mh);
}

SPIDERH_ptr PyEMData::py_getSpiHed()
{
    SPIDERH* sh = getSpiHed();
    return SPIDERH_ptr(sh);
}

  
PyEMData_ptr PyEMData::py_doFFT()
{
    PyEMData* em = static_cast<PyEMData*> (doFFT());
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_doIFT()
{
    PyEMData* em = static_cast<PyEMData*> (doIFT());
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_doRadon()
{
    PyEMData* em = static_cast<PyEMData*> (doRadon());
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_vertACF(int maxdy)
{
    PyEMData* em = static_cast<PyEMData*> (vertACF(maxdy));
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_convolute(PyEMData& with)
{
    PyEMData* em = static_cast<PyEMData*> (convolute(with.deref()));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_calcCCF(PyEMData& with, PyEMData& filter)
{
    PyEMData* em = static_cast<PyEMData*> (calcCCF(with.deref(), filter.deref()));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_calcCCFX(PyEMData& with, int y0, int y1, int nosum)
{
    PyEMData* em = static_cast<PyEMData*> (calcCCFX(with.deref(), y0, y1, nosum));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_calcMCF(PyEMData& with, PyEMData& filter)
{
    PyEMData* em = static_cast<PyEMData*> (calcMCF(with.deref(), filter.deref()));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_littleBigDot(PyEMData& with, int sigma)
{
    PyEMData* em = static_cast<PyEMData*> (littleBigDot(with.deref(), sigma));
    return PyEMData_ptr(em);
}

float PyEMData::py_transAlign(PyEMData& with, int useparent, int intonly,
			      int maxshift, Float& peak)
{
    float* p_peak = deref(peak);
    return transAlign(with.deref(), useparent, intonly, maxshift, p_peak);    
}

float PyEMData::py_rotAlign(PyEMData& with)
{
    return rotAlign(with.deref());
}

float PyEMData::py_rotAlignPrecen(PyEMData& with)
{
    return rotAlignPrecen(with.deref());
}


float PyEMData::py_rotAlignCH(PyEMData& with, int irad, int orad)	
{
    return rotAlignCH(with.deref(), irad, orad);
}

float PyEMData::py_transAlign3d(PyEMData& with, int useparent, int intonly)	
{
    return transAlign3d(with.deref(), useparent, intonly);
}


PyEMData_ptr PyEMData::py_RTFAlignRadon(PyEMData& with, int maxshift,
					PyEMData& thisf, PyEMData& radonwith,
					PyEMData& radonthis, PyEMData& radonthisf)
{
    EMData* d = RTFAlignRadon(with.deref(), maxshift, thisf.deref(),
			      radonwith.deref(), radonthis.deref(), radonthisf.deref());
    PyEMData* em = static_cast<PyEMData*> (d);
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_RTAlignRadon(PyEMData& with, int maxshift,
				       PyEMData& radonwith, PyEMData& radonthis)
{
    EMData* d = RTAlignRadon(with.deref(), maxshift, radonwith.deref(), radonthis.deref());
    PyEMData* em = static_cast<PyEMData*> (d);
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_RTFAlignSlow(PyEMData& with, PyEMData& flip,
				       int maxshift)
{
    EMData* d = RTFAlignSlow(with.deref(), flip.deref(), maxshift);
    PyEMData* em = static_cast<PyEMData*>(d); 
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_RTFAlignSlowest(PyEMData& with, PyEMData& flip,
					  int maxshift)
{
    EMData* d = RTFAlignSlowest(with.deref(), flip.deref(), maxshift);
    PyEMData* em = static_cast<PyEMData*>(d);
    return PyEMData_ptr(em);
}


PyEMData_ptr PyEMData::py_RTFAlign(PyEMData& with, PyEMData& flip,
				   int usedot, int maxshift)
{
    EMData* d = RTFAlign(with.deref(), flip.deref(), usedot, maxshift);
    PyEMData* em = static_cast<PyEMData*>(d);
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_RFAlign(PyEMData& with, PyEMData& flip, int imask)
{
    PyEMData* em = static_cast<PyEMData*> (RFAlign(with.deref(), flip.deref(), imask));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_RTAlign(PyEMData& with, int usedot, int maxshift)
{
    PyEMData* em = static_cast<PyEMData*> (RTAlign(with.deref(), usedot,maxshift));
    return PyEMData_ptr(em);
}


void PyEMData::py_refineAlign(PyEMData& with, int mode)
{
    refineAlign(with.deref(), mode);
}

PyEMData_ptr PyEMData::py_makeRFP(int premasked, int unwrap)
{
    PyEMData* em = static_cast<PyEMData*> (makeRFP(premasked, unwrap));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_unwrap(int r1, int r2, int xs,
				 int dx, int dy, int do360)   
{
    PyEMData* em = static_cast<PyEMData*> (unwrap(r1, r2, xs, dx, dy, do360));
    return PyEMData_ptr(em);
}

void PyEMData::py_calcRCF(PyEMData& with, PyList sum_list, int NS)
{
#ifdef DEBUG
    assert(sum_list.size() >= NS);
#endif
    
    float* sum_array = new float[sum_list.size()];
    list2array<float>(sum_list, sum_array);    
    calcRCF(with.deref(), sum_array, NS);
    array2list(sum_array, sum_list);
    delete [] sum_array;
    sum_array = 0;
}

void PyEMData::py_applyRadFn(int n, float x0, float dx, PyList y_list, int interp)
{
#ifdef DEBUG
    assert(y_list.size() >= n);
#endif
    
    float* y_array = new float[y_list.size()];
    list2array<float>(y_list, y_array);
    applyRadFn(n, x0, dx, y_array, interp);
    array2list<float>(y_array, y_list);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_addRadNoise(int n, float x0, float dx, PyList y_list, int interp)	   
{
#ifdef DEBUG
    assert(y_list.size() >= n);
#endif
    
    float* y_array = new float[y_list.size()];
    list2array<float>(y_list, y_array);
    addRadNoise(n, x0, dx, y_array, interp);
    array2list<float>(y_array, y_list);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_subtract(PyEMData& data)	
{
    subtract(data.deref());
}


void PyEMData::py_normalizeTo(PyEMData& noisy, int keepzero,
			      int invert, Float& mult, Float& add)	
{
    float* p_mult = deref(mult);
    float* p_add = deref(add);
    normalizeTo(noisy.deref(), keepzero, invert, p_mult, p_add);
}


void PyEMData::py_leastSquareNormalizeTo(PyEMData& to,  float low_thresh,
					 float high_thresh, Float& scale,
					 Float& shift)
{
    float* p_scale = deref(scale);
    float* p_shift = deref(shift);
    leastSquareNormalizeTo(to.deref(), low_thresh, high_thresh, p_scale, p_shift);
}

void PyEMData::py_maskNormalize(PyEMData& mask, int sigmatoo) 
{
    maskNormalize(mask.deref(), sigmatoo);
}

void PyEMData::py_calcRadDist(int n, float x0, float dx, PyList y_list)
{
#ifdef DEBUG
    assert(y_list.size() >= n);
#endif
    
    float* y_array = new float[y_list.size()];
    list2array(y_list, y_array);
    calcRadDist(n, x0, dx, y_array);
    array2list(y_array, y_list);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_calcRadDist(int n, float x0, float dx, PyList y_list,
			      float acen, float amwid)
{
#ifdef DEBUG
    assert(y_list.size() >= 0);
#endif
    
    float* y_array = new float[y_list.size()];
    list2array<float>(y_list, y_array);
    calcRadDist(n, x0, dx, y_array, acen, amwid);
    array2list<float>(y_array, y_list);
    delete [] y_array;
    y_array = 0;
}

void PyEMData::py_calcAzDist(int n, float a0, float da, PyList d_list,
			     float rmin, float rmax)
{
#ifdef DEBUG
    assert(d_list.size() >= n);
#endif
    
    float* d_array = new float[d_list.size()];
    list2array(d_list, d_array);
    calcAzDist(n, a0, da, d_array, rmin, rmax);
    array2list(d_array, d_list);
    delete [] d_array;
    d_array = 0;
}

float PyEMData::py_dot(PyEMData& data, int evenonly)
{
    return dot(data.deref(), evenonly);
}

float PyEMData::py_lcmp(PyEMData& noisy, int keepzero, int invert, 
			Float& scale, Float& shift)
{
    float* p_scale = deref(scale);
    float* p_shift = deref(shift);    
    return lcmp(noisy.deref(), keepzero, invert, p_scale, p_shift);
}

float PyEMData::py_pcmp(PyEMData& data, XYData& snr)
{
    return pcmp(data.deref(), deref(snr));
}


PyList PyEMData::py_ctfCurve(int type, XYData& sf)
{
    float* sf_array = ctfCurve(type, deref(sf));
    // convert sf-array to a list????
    return PyList();
}

void PyEMData::py_ctfMap(int type, XYData& sf)
{
    ctfMap(type, deref(sf));
}


float PyEMData::py_fscmp(PyEMData& with)
{
    return fscmp(with.deref());
}

PyList PyEMData::py_fsc(PyEMData& with)
{
    // return NULL or array: size = ?
    float* array = fsc(with.deref());
    return PyList();
    
}



void PyEMData::py_makeAverageWT(PyList in_pylist, PyEMData& curves, XYData& sf)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageWT(in_emlist, curves.deref(), deref(sf));

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageCTFCW(PyList in_pylist, XYData& sf, PyList SNR_list)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);
    
    int sz = SNR_list.size();
    float* SNR_array = new float[sz];
    list2array(SNR_list, SNR_array);

    makeAverageCTFCW(in_emlist, deref(sf), SNR_array);

    array2list(SNR_array, SNR_list);
    delete [] SNR_array;
    SNR_array = 0;

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverage(PyList in_pylist, PyEMData& sigma)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverage(in_emlist, sigma.deref());

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageIter(PyList in_pylist)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);
    makeAverageIter(in_emlist);

    delete in_emlist;
    in_emlist = 0;
}


void PyEMData::py_makeAverageCTFC(PyList in_pylist, float filter)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageCTFC(in_emlist, filter);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_makeAverageCTFCWauto(PyList in_pylist, const char *outfile)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeAverageCTFCWauto(in_emlist, const_cast<char*>(outfile));
    
    delete in_emlist;
    in_emlist = 0;
}


void PyEMData::py_makeMedian(PyList in_pylist)
{
    List* in_emlist = new List();
    pylist2emlist(in_pylist, in_emlist);

    makeMedian(in_emlist);

    delete in_emlist;
    in_emlist = 0;
}

void PyEMData::py_add(PyEMData& in)
{
    add(in.deref());
}

void PyEMData::py_addIncoherent(PyEMData& in)						
{
    addIncoherent(in.deref());
}

void PyEMData::py_mult(PyEMData& in, int recip)
{
    mult(in.deref(), recip);
}

void PyEMData::py_commonLines(PyEMData& data1, PyEMData& data2, int mode, 
			      int steps, int horiz)
{
    commonLines(data1.deref(), data2.deref(), mode, steps, horiz);
}

void PyEMData::py_commonLinesR(PyEMData& data1, PyEMData& data2, 
			       int steps, int horiz)				
{
    commonLinesR(data1.deref(), data2.deref(), steps, horiz);
}

PyEMData_ptr PyEMData::py_project3d(float alt, float az,  float phi, 
				    int mode,  float thr)
{
    PyEMData* em = static_cast<PyEMData*> (project3d(alt, az, phi,mode, thr));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_fftSlice(float alt, float az, float phi, int mode)
{
    PyEMData* em = static_cast<PyEMData*> (fftSlice(alt, az, phi, mode));
    return PyEMData_ptr(em);
}

void PyEMData::py_interpFT3d(float x, float y, float z, PyList ret_list, int mode)
{
    // only 2 elements are used in ret_list
#ifdef DEBUG
    assert(ret_list.size() >= 2);
#endif
    float* ret_array =  new float[ret_list.size()];
    list2array(ret_list, ret_array);
    interpFT3d(x, y, z, ret_array, mode);
    array2list(ret_array, ret_list);
    delete [] ret_array;
    ret_array = 0;
}


float PyEMData::py_normSlice(PyEMData& slice, float alt, float az,
			     float phi, Float& phaseresid)
{
    float* p_phaseresid = deref(phaseresid);
    return normSlice(slice.deref(), alt, az, phi, p_phaseresid);
}


void PyEMData::py_insertSlice(PyEMData& slice, float alt, float az,
			      float phi, int mode, float weight)
{
    insertSlice(slice.deref(), alt, az, phi, mode, weight);
}


void PyEMData::py_insertSliceWF(PyEMData& slice, float alt, float az,
				float phi, int mode, PyList SNR_list,
				float padratio)
{
    float* SNR_array = new float[SNR_list.size()];
    list2array(SNR_list, SNR_array);
    insertSliceWF(slice.deref(), alt, az, phi, mode, SNR_array, padratio);
    array2list(SNR_array, SNR_list);
    delete [] SNR_array;
    SNR_array = 0;
}

void PyEMData::py_insertSlice3DReal(PyEMData& slice, float weight)

{
    insertSlice3DReal(slice.deref(), weight);
}


PyList PyEMData::py_getCTF()
{
    // length = 11
    const int SZ = 11;
    float* ctf_array = getCTF();
    PyList ctf_list(SZ);
    array2list(ctf_array, ctf_list);
    return ctf_list;
}

void PyEMData::py_setCTF(PyList ctf_list)
{
    // length = 11
    const int SZ = 11;
#ifdef DEBUG
    assert(ctf_list.size() == SZ);
#endif
    float* ctf_array = new float[SZ];
    list2array<float>(ctf_list, ctf_array);
    setCTF(ctf_array);
    delete [] ctf_array;
    ctf_array = 0;
}

int PyEMData::py_CTFCmp(PyEMData& with)
{
    return CTFCmp(with.deref());
}

PyEMData_ptr PyEMData::py_iftSlice()
{
    PyEMData* em = static_cast<PyEMData*> (iftSlice());
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_row(int n)
{
    PyEMData* em = static_cast<PyEMData*> (row(n));
    return PyEMData_ptr(em);
}

PyEMData_ptr PyEMData::py_col(int n)
{
    PyEMData* em = static_cast<PyEMData*> (col(n));
    return PyEMData_ptr(em);
}

void PyEMData::py_setRow(PyEMData& d, int n)
{
    setRow(d.deref(), n);
}

void PyEMData::py_setCol(PyEMData& d, int n)
{
    setCol(d.deref(), n);
}

PyEMData_ptr PyEMData::py_mxMult(PyEMData& second)
{
    PyEMData* em = static_cast<PyEMData*> (mxMult(second.deref()));
    return PyEMData_ptr(em);
}


float PyEMData::py_dist(PyEMData& second, int n)
{
    return dist(second.deref(), n);
}



enum EMDataState {
    EMDATA_FLAGS,
    EMDATA_DATA,
    EMDATA_NX,
    EMDATA_NY,
    EMDATA_NZ,
    EMDATA_DX,
    EMDATA_DY,
    EMDATA_DZ,
    EMDATA_EULER,
    EMDATA_NIMG,
    EMDATA_PIXEL,
    EMDATA_CTF,
    EMDATA_MIN, 
    EMDATA_MAX,
    EMDATA_MINLOC,
    EMDATA_MAXLOC,
    EMDATA_MEAN,
    EMDATA_SIGMA,
    EMDATA_SUMSQ,
    EMDATA_MEAN2,
    EMDATA_SIGMA2,
    EMDATA_DICT,
    NUM_EMDATA_STATES
};


python::ref EMData_getstate(const python::tuple& args,
			    const python::dictionary& keywords)
{
    check_args(args, keywords, 1);

    PyEMData& em = BOOST_PYTHON_CONVERSION::from_python(args[0].get(),
							python::type<PyEMData&>());
    
    python::tuple result(NUM_EMDATA_STATES);
    result.set_item(EMDATA_FLAGS, em.getFlags());

    PyObject* num_array = em.get_num_array();
    result.set_item(EMDATA_DATA, num_array);

    
    result.set_item(EMDATA_NX, em.xSize());
    result.set_item(EMDATA_NY, em.ySize());
    result.set_item(EMDATA_NZ, em.zSize());
    result.set_item(EMDATA_DX, em.Dx());
    result.set_item(EMDATA_DY, em.Dy());
    result.set_item(EMDATA_DZ, em.Dz());

    PyEuler euler = em.py_getEuler();
    result.set_item(EMDATA_EULER, euler);

    result.set_item(EMDATA_NIMG, em.NImg());
    result.set_item(EMDATA_PIXEL, em.Pixel());

    float* ctf_array = em.getCTF();
    PyList ctf_list(11);
    array2list<float>(ctf_array, ctf_list);    
    result.set_item(EMDATA_CTF, ctf_list);
    
    result.set_item(EMDATA_MIN, em.Min());
    result.set_item(EMDATA_MAX, em.Max());
    result.set_item(EMDATA_MINLOC, em.MinLoc());
    result.set_item(EMDATA_MAXLOC, em.MaxLoc());
    result.set_item(EMDATA_MEAN, em.Mean());
    result.set_item(EMDATA_SIGMA, em.Sigma());
    result.set_item(EMDATA_SUMSQ, em.SumSq());
    result.set_item(EMDATA_MEAN2, em.Mean2());
    result.set_item(EMDATA_SIGMA2, em.Sigma2());

    
    python::ref mydict = python::getattr(args[0], "__dict__");
    result.set_item(EMDATA_DICT, mydict);

    
    return result.reference();
}

PyObject* EMData_setstate(const python::tuple& args,
			  const python::dictionary& keywords)
{
    set_dict(args, keywords, NUM_EMDATA_STATES);
    
    PyEMData& em = BOOST_PYTHON_CONVERSION::from_python(args[0].get(),
							python::type<PyEMData&>());
    python::tuple state = BOOST_PYTHON_CONVERSION::from_python(args[1].get(),
							       python::type<python::tuple>());
    

    int flags = get_int(state[EMDATA_FLAGS]);
    em.setFlags(flags);

    int nx = get_int(state[EMDATA_NX]);
    int ny = get_int(state[EMDATA_NY]);
    int nz = get_int(state[EMDATA_NZ]);    
    em.setSize(nx, ny, nz);
    
    PyObject* num_array = BOOST_PYTHON_CONVERSION::from_python(state[EMDATA_DATA].get(),
							       python::type<PyObject*>());
    em.set_num_array(num_array, nx*ny*nz);
    
    
    float dx = get_float(state[EMDATA_DX]);
    float dy = get_float(state[EMDATA_DY]);
    float dz = get_float(state[EMDATA_DZ]);    
    em.setTAlign(dx, dy, dz);    
    
    PyEuler euler = BOOST_PYTHON_CONVERSION::from_python(state[EMDATA_EULER].get(),
							 python::type<PyEuler>());
    em.py_setRAlign(euler);
    
    int nimg = get_int(state[EMDATA_NIMG]);
    em.setNImg(nimg);

    float pixel = get_float(state[EMDATA_PIXEL]);
    em.setPixel(pixel);
    
    PyList ctf_list = BOOST_PYTHON_CONVERSION::from_python(state[EMDATA_CTF].get(),
							   python::type<PyList>());
#ifdef DEBUG
    assert(ctf_list.size() == 11);
#endif
    float* ctf_array = new float[ctf_list.size()];
    list2array<float>(ctf_list, ctf_array);
    em.setCTF(ctf_array);
    delete [] ctf_array;
    ctf_array = 0;

    float min = get_float(state[EMDATA_MIN]);
    em.setMin(min);

    float max = get_float(state[EMDATA_MAX]);
    em.setMax(max);

    int minloc = get_int(state[EMDATA_MINLOC]);
    em.setMinLoc(minloc);

    int maxloc = get_int(state[EMDATA_MAXLOC]);    
    em.setMaxLoc(maxloc);

    float mean = get_float(state[EMDATA_MEAN]);
    em.setMean(mean);

    float sigma = get_float(state[EMDATA_SIGMA]);
    em.setSigma(sigma);
    
    float sumsq = get_float(state[EMDATA_SUMSQ]);
    em.setSumSq(sumsq);
    
    float mean2 = get_float(state[EMDATA_MEAN2]);
    em.setMean2(mean2);

    float sigma2 = get_float(state[EMDATA_SIGMA2]);
    em.setSigma2(sigma2);

    
    return python::detail::none();
}


const char* PyEMData::py_doc(const char* funcname)
{
    if (!funcname) {
	return "";
    }

    static bool has_built = false;
    static python::dictionary dict = python::dictionary();
    
    if (!has_built) {
	dict["EMData"] = "EMData() \n  construct an empty EMData instance";
	dict["get_num_array"] = "Numeric.array get_num_array() \n return the EM data in a numeric array";

	dict["readImage"] = "int readImage(string filename, int imagetype, int n, int nodata) \n  return 0 if succeeds; -1 if fails.";

	dict["writeImage"] = "int writeImage(string filename, int image_type, int n, int nolock, int mode) \n  return 0 if succeeds; -1 if fails.";
	
	dict["copy"] = "EMData copy(int withfft=0, int withparent=1)\n  returns an exact copy of itself";

	dict["copyHead"] = "EMData copyHead() \n copies descriptive info, but not data";
	dict["clip"] = "EMData clip(int x0, int y0, int w, int h) \nEMData clip(int x0,int y0,int z0,int w,int h,int d) \n  inclusive clip. Pads if larger than data";

	dict["insertClip"] = "insertClip(EMData block, int x0, int y0) \n used for 2D only";
	dict["serialOut"] = "serialOut(int filedes) \n  used ONLY for FILESERVER operation, not safe otherwise";
	dict["calcHist"] = "calcHist(list hist, int n, float hmin=0, float hmax=0,int add=0)";
	dict["update"] = "update()";
	
	dict["setPixel"] = "setPixel(float pix)";
	dict["Pixel"] = "float Pixel()";
	dict["Min"] = "float Min()";
	dict["Max"] = "float Max()";
	dict["MinLoc"] = "int MinLoc()";
	dict["MaxLoc"] = "int MaxLoc()";

	dict["Mean"] = "float Mean()";
	dict["Sigma"] = "float Sigma()";
	dict["SumSq"] = "float SumSq()";
	dict["Mean2"] = "float Mean2()";
	dict["Sigma2"] = "float Sigma2()";
	dict["edgeMean"] = "float edgeMean()";
	dict["cirMean"] = "float cirMean()";
	
	
	dict["setParent"] = "setParent(EMData data)";
	dict["Parent"] = "EMData Parent()";
	
	dict["setName"] = "setName(string name)";
	dict["Name"] = "string Name()";
	dict["setPath"] = "setPath(string path)";
	dict["Path"] = "string Path()";
	
	dict["setPathN"] = "setPathN(int n)";
	
	dict["PathN"] = "int PathN()";
	dict["xSize"] = "int xSize()";
	dict["ySize"] = "int ySize()";
	dict["zSize"] = "int zSize()";
	dict["setSize"] = "int setSize(int x,int y=1, int z=1)";
	
	dict["setRI"] = "setRI(int i)";
	dict["setComplex"] = "setComplex(int i)";
	dict["setComplexX"] = "setComplexX(int i)";

	dict["isComplex"] = "int isComplex()";
	dict["isComplexX"] = "int isComplexX()";
	dict["hasCTF"] = "int hasCTF()";
	dict["hasCTFF"] = "int hasCTFF()";
	dict["getFlags"] = "int getFlags()";
	dict["isFlipped"] = "int isFlipped()";

	dict["setFlipped"] = "setFlipped(int f)";
	
	dict["aliPeak"] = "float aliPeak()";
	dict["Dx"] = "float Dx()";
	dict["Dy"] = "float Dy()";
	dict["Dz"] = "float Dz()";
	
	dict["alt"] = "float alt()";
	dict["az"] = "float az()";
	dict["phi"] = "float phi()";
	
	dict["getEuler"] = "Euler getEuler()";
	dict["setTAlign"] = "setTAlign(float x,float y,float z=0)";
	dict["setRAlign"] = "setRAlign(float Alt,float Az=0,float Phi=0)\nsetRAlign(Euler angle)";

	dict["setNImg"] = "setNImg(int n)";
	dict["NImg"] = "int NImg()";
	dict["getATime"] = "long getATime()";
	
	dict["getImagicHed"] = "imagicH getImagicHed()";
	dict["setImagicHed"] = "setImagicHed(imagicH hed)";
	dict["getMRCHed"] = "mrcH getMRCHed()";	
	dict["getSpiHed"] = "SPIDERH getSpiHed()";
	
	dict["zero"] = "zero()";
	dict["one"] = "one()";
	
	//dict["renderAmp8"] = "int renderAmp8(unsigned char * data , int x , int y, int xsize, int ysize, int bpl, float scale, int mingray, int maxgray,float renMin, float renMax)";
	//dict["renderPha24"] = "int renderPha24(unsigned char * data , int x , int y , int xsize , int ysize , int bpl , float scale, float renMin, float renMax)";

	dict["ri2ap"] = "ri2ap() \n  converts real/imaginary to amp/pha (complex only)";
	dict["ap2ri"] = "ap2ri() \n  converts amp/pha to real/imaginary (complex only)";
	dict["doFFT"] = "EMData doFFT()";
	dict["doIFT"] = "EMData doIFT()";
	dict["gimmeFFT"] = "gimmeFFT() \n  gives the caller ownership of the recently obtained fft so it won't be freed automatically";
	dict["doRadon"] = " EMData doRadon() \n  Radon transform of image, comes out square";
	dict["vertACF"] = "EMData vertACF(int maxdy) \n  vertical autocorrelation calculated in real-space over a range of translations";
	dict["doDWT"] = "doDWT(int basis, int level) \n  in place DWT, basis/level defined elsewhere";
	dict["doIWT"] = "doIWT(int basis, int level) \n  inverse wavelet x-form";
	
	dict["DWTFilt"] = "DWTFilt(int basis, int level, float thresh) \n  in place nonlinear threshold wavelet filter. thresh is with respect to the 'ideal' threshold";
	dict["convolute"] = "EMData convolute(EMData with)";
	dict["calcCCF"] = "EMData calcCCF(EMData with,EMData filter=None) \n  cross correlation between 2 images";
	dict["calcCCFX"] = "EMData calcCCFX(EMData with,int y0=0,int y1=-1,int nosum=0) \n  calculates sum of x-axis CCFs";
	dict["calcMCF"] = "EMData calcMCF(EMData with,EMData filter=None) \n  mutual correlation between 2 images";
	dict["littleBigDot"] = "EMData littleBigDot(EMData with,int sigma=0) \n  CCF in real-space, designed for cases where with is small";
	dict["transAlign"] = "float transAlign(EMData with,int useparent,int intonly,int maxshift, Float peak=0)  \n  translational alignment using CCF";
	dict["rotAlign"] = "float rotAlign(EMData with) \n  rotational alignment using angular correlation";
	dict["rotAlignPrecen"] = "float rotAlignPrecen(EMData with) \n  rotational alignment assuming centers are correct";
	dict["rotAlignCH"] = "float rotAlignCH(EMData with,int irad=8,int orad=0) \n  rotational alignment via circular harmonic";
	dict["transAlign3d"] = "float transAlign3d(EMData with,int useparent=0,int intonly=0) \n  translational alignment using CCF";
	dict["RTFAlignRadon"] = "EMData RTFAlignRadon(EMData with,int maxshift=-1,EMData thisf=None,EMData radonwith=None ,EMData radonthis=None,EMData radonthisf=None) \n  rotational, translational and flip alignment with Radon transforms";

	dict["RTAlignRadon"] = "EMData RTAlignRadon(EMData with,int maxshift=-1,EMData radonwith=None,EMData radonthis=None)";
	dict["RTFAlignSlow"] = "EMData RTFAlignSlow(EMData with,EMData flip=None,int maxshift=-1)\n  rotational, translational and flip alignment using real-space methods, SLOW";
	dict["RTFAlignSlowest"] = "EMData RTFAlignSlowest(EMData with,EMData flip=None,int maxshift=-1)\n  rotational, translational and flip alignment using exhaustive search. VERY SLOW";
	dict["RTFAlign"] = "EMData RTFAlign(EMData with=None,EMData flip=None,int usedot=1,int maxshift=-1) \n  rotational, translational and flip alignment";
	dict["RFAlign"] = "EMData RFAlign(EMData with,EMData flip=None,int imask=0)\n  rotational and flip alignment";
	dict["refineAlign"] = "refineAlign(EMData with,int mode=1)";
	dict["makeRFP"] = "EMData makeRFP(int premasked=0,int unwrap=1)\n  Makes a 'rotational footprint', which is an 'unwound' autocorrelation function. if unwrap is 0 the rfp is not cached. generally the image should be edgenormalized and masked before using this";
	dict["unwrap"] = "EMData unwrap(int r1=-1,int r2=-1,int xs=-1,int dx=0,int dy=0,int do360=0)\n  maps polar coordinates to cylindrical coordinates";
	dict["calcRCF"] = "calcRCF(EMData with, list sum,int NS)";
	dict["rotateX"] = "rotateX(int dx) \n  This performs a translation of each line along x with wraparound; this is equivalent to a rotation when performed on 'unwrapped' maps";
	dict["rotateAndTranslate"] = "rotateAndTranslate(float scale=1.0,float dxc=0,float dyc=0,float dzc=0,int r=0)";
	dict["rotateAndTranslateFast"] = "rotateAndTranslateFast(float scale=1.0)\n  rotate and translate using current settings; behavior changes if parent is/not set";
	dict["fastTranslate"] = "fastTranslate(int inplace=1)\n  fast integer-only translation";
	dict["rot180"] = "rot180()\n  fast rotation by 180 degrees";
	dict["cmCenter"] = "cmCenter(int intonly=1)\n  center at center of mass, ignores old dx,dy";
	dict["applyMask"] = "applyMask(int r,int type)\n  applies a circular mask to the data. type=0 is a step cutoff to the mean value; type=4 is a step cutoff to zero; type=1 fills in with flatband random noise";
	dict["autoMask"] = "autoMask(float thresh,float filter=.1) \n  Attempts to automatically mask out the particle, excluding other particles in the box, etc. thresh runs from ~ -2 to 2, negative numbers for dark. protien and positive numbers for light protein (stain) filter is expressed as a fraction of the Fourier radius ";

	dict["applyRadFn"] = "applyRadFn(int n,float x0,float dx, list y,int interp=1)";
	dict["addRadNoise"] = "addRadNoise(int n,float x0,float dx, list y,int interp)";
	dict["subtract"] = "subtract(EMData data) \n  subtracts 'data' from the current set";
	dict["vFlip"] = "vFlip()\n  flips image vertically";
	dict["hFlip"] = "hFlip()\n  flips image horizontally";
	dict["normalize "] = "normalize()\n  mean -> 0, std dev -> 1";
	dict["normalizeTo"] = "normalizeTo(EMData noisy,int keepzero=0,int invert=0, Float mult=0, Float add=0)\n  This will multiply 'this' by a constant so. it is scaled to the signal in 'to'keepzero will exclude zero values, and keep them at zero in the result";
	dict["normalizeMax"] = "normalizeMax()\n  mean -> 0, std dev -> 1";
	dict["edgeNormalize"] = "edgeNormalize(int cir=0)\n  normalizes so mean -> value at edge of image, if cir=1, uses 2 pixel circular border";
	dict["xNormalize"] = "xNormalize(int mode=0)\n  normalizes each row in the image individually";
	dict["addMaskShell"] = "addMaskShell(int shells)\n  Will add additional shells/rings to an existing 1/0 mask image";
	dict["leastSquareNormalizeTo"] = "leastSquareNormalizeTo(EMData to, float low_thresh=FLT_MIN, float high_thresh=FLT_MAX, Float scale=0, Float shift=0)";
	dict["maskNormalize"] = "maskNormalize(EMData mask,int sigmatoo=1)\n  Uses a 1/0 mask defining a region to use for the zero-normalization, if sigma is 0, std-dev not modified";
	dict["invert"] = "invert()\n  multiply by -1";
	dict["calcRadDist"] = "calcRadDist(int n,float x0,float dx, list y)";
	dict["calcRadDist"] = "calcRadDist(int n,float x0,float dx, list y,float acen,float amwid)\n  calculates radial distribution, works for real and imaginary images. x must be defined on input and y must be allocated. amin,amax are in radians";
	dict["calcAzDist"] = "calcAzDist(int n,float a0,float da, list data,float rmin,float rmax)";
	dict["radialAverage"] = "radialAverage()\n  makes image circularly symmetric";
	dict["radialSubtract"] = "radialSubtract()\n  subtracts circularly symmetric part of image";	
	
	dict["toCorner"] = "toCorner()\n  Translates a centered image to the corner";
	dict["dot"] = "float dot(EMData data,int evenonly=0)\n  dot product of 2 images (same size !)";

	dict["lcmp"] = "float lcmp(EMData noisy,int keepzero=0,int invert=0,Float scale=0,Float shift=0)\n  linear comparison of 2 data sets (smaller better)";
	dict["pcmp"] = "float pcmp(EMData data,XYData snr=0)\n  amplitude weighted mean phase error, 'data' should be less noisy than 'this'";
	dict["fscmp"] = " float fscmp(EMData with)\n  returns a quality factor based on FSC between images larger numbers indicate a better match";
	dict["fsc"] = "list fsc(EMData with)\n  returns the Fourier ring/shell correlation coef.as an array with ysize/2 elements (corners not calculated) the calling process must free the array";
	dict["dCenter"] = "float dCenter()\n  This will find the density value at the peak of the image histogram, sort of like the mode of the density";

	
	dict["sigDiff"] = "float sigDiff()\n  Calculates sigma above and below the mean and returns the difference between them";
	dict["realFilter"] = "realFilter(int type, float val1=0,float val2=0, float val3=0)\n  a variety of real-space filters";

	dict["filter"] = "filter(float highpass,float lowpass,int type)\n  fourier filters an image, real or imaginary";
	dict["makeAverage"] = "makeAverage(list in,EMData sigma=None)\n  averages the images in the list, result in 'this' optionally makes a sigma image as well";
	dict["makeAverageIter"] = "makeAverageIter(list in)";

	dict["makeAverageWT"] = "makeAverageWT(list in,EMData curves,XYData sf)\n  averages the images in the list with SNR weightingbut no CTF correction";

	dict["makeAverageCTFC"] = "makeAverageCTFC(list in,float filtr)\n  averages the images in the list with CTF correction filtr is in angstroms,";

	dict["makeAverageCTFCWauto"] = "makeAverageCTFCWauto(list in, string outfile=None)\n  This does CTF correction with a Wiener filter where the filter is estimated directly from the data";

	dict["makeAverageCTFCW"] = "makeAverageCTFCW(list in,XYData sf, list SNR=None)\n  averages the images in the list with CTF correction";
	
	dict["makeMedian"] = "makeMedian(list in)\n  Median of the images in the list, result in 'this'";
	dict["add"] = "add(EMData in)\n  adds 'in' to 'this'";
	dict["addIncoherent"] = "addIncoherent(EMData in)\n  adds 'in' to 'this'";
	dict["mult"] = "mult(EMData in,int recip=0)\n  multiplies 'in' by 'this', if recip, then divide";
	dict["multConst"] = "multConst(float)\n  multiply by constant";
	
	dict["addConst"] = "addConst(float)\n add a constant";

	dict["meanShrink"] = " meanShrink(int i)\n  reduces the size of the image by a factor of i, using the average value of the pixels in a block";

	dict["medianShrink"] = "medianShrink(int i)\n  reduces the size of the image by a factor of i using a local median filter";


	dict["commonLines"] = "commonLines(EMData d1,EMData d2,int mode=0,int steps=180,int horiz=0) \n  finds common lines between 2 complex images. mode 0 is a summed dot-product; mode 1 is weighted phase residual;steps is 1/2 the resolution of the map";
	
	dict["commonLinesR"] = "commonLinesR(EMData data1,EMData data2,int steps=180,int horiz=0)";
	dict["project3d"] = "EMData project3d(float alt,float az, float phi, int mode, float thr=1.0) \n  makes a 3d projection of a volume, mode=-1 does real space projection, 1-5 use fftslice for fourier projection";
	dict["fftSlice"] = "EMData fftSlice(float alt,float az,float phi,int mode=5)\n  takes a slice from a 3d complex image for generating projections.";
	
	dict["setup4Slice"] = "setup4Slice(int redo=1)";
	dict["interpFT3d"] = "interpFT3d(float x,float y,float z, list ret,int mode)";
	dict["setup4IS"] = " setup4IS(int size)";

	dict["normSlice"] = "normSlice(EMData slice,float alt,float az,float phi,Float phaseresid=0)\n  try to make sure the slice is normalized properly";
	dict["insertSlice"] = "insertSlice(EMData slice,  float alt,float az,float phi,int mode=5,float weight=1.0) \n  this inserts a complex slice into an image that has been setup4IS()";
	dict["insertSliceWF"] = "insertSliceWF(EMData slice,  float alt,float az,float phi,int mode,list SNR,float padratio)\n  this inserts a complex slice into an image that has been setup4IS()";
	dict["doneSlice"] = "doneSlice(int log=0)\n   once all of the slices have been added with insertSlice(), this routine does a normalization and returns a real 3d volume";

	dict["doneSliceWF"] = "doneSliceWF(int log=0)\n  It should only be used in conjunction with insertSliceWF";
	
	dict["insertSlice3DReal"] = "insertSlice3DReal(EMData slice,float weight)";
	dict["ctfCurve"] = "list ctfCurve(int type=0,XYData sf=None)\n  calculates a CTF curve optional structure factor data for true SNR calc";
	dict["ctfMap"] = "ctfMap(int type,XYData sf=None)";
	dict["getCTF"] = "list getCTF()";
	dict["setCTF"] = "setCTF(list c)";
	dict["setCTFF"] = "setCTFF(string s)";
	dict["clearCTF"] = "clearCTF()";
	dict["CTFCmp"] = "int CTFCmp(EMData with)\n  used to detect if 2 images have the same CTF parameters";
	dict["subNoise"] = "subNoise()";
	dict["iftSlice"] = "EMData iftSlice()\n  This does an IFT of a volume reconstructed from slices. Includes a real-space correction for the linear interpolation in Fourier space";
	dict["valueAt"] = "float valueAt(int x, int y, int z) \nfloat valueAt(int x, int y)";
	dict["valueAtSafe"] = "float valueAtSafe(int x, int y, int z) \nfloat valueAtSafe(int x, int y)";
	
	dict["setValueAt"] = "setValueAt(int x, int y, int z, float v)\n it does not set the changed flag";
	dict["row"] = "EMData row(int n)\n  extract vector from matrix";
	dict["col"] = "EMData col(int n)\n  extract vector from matrix";
	dict["setRow"] = "setRow(EMData d,int n)\n  insert vector into matrix";
	dict["setCol"] = "setCol(EMData d,int n)\n  insert vector into matrix";
	dict["mxMult"] = " EMData mxMult(EMData second)";
	dict["dist"] = "float dist(EMData second, int n=0)";
	

	
	has_built = true;
    }
    
    const char* fdoc = BOOST_PYTHON_CONVERSION::from_python(dict.get_item(funcname).get(),
							    python::type<const char*>());
    if (!fdoc) {
	fdoc = "Error: not defined";
    }
    return fdoc;
}


