// align2d.C 8/99  Steve Ludtke
// Copyright 1999 Baylor College of Medicine

//N align2d
//F This program will align a set of images to a reference
//T LA
//1
//P <fixed input>	Reference image
//P <input>	Images to be aligned
//P <output>	Output filename
//P [fixn=<n>]	Image number to use in fixed input stack
//P [noflip]	Rotational/translational alignment without flips
//P [precen]	No translational alignment
//P [seqali]	Align 1st image to ref, 2nd to 1st, 3rd to 2nd, etc.
//P [eulali]	Align each image to images with similar Eulers, referring to the 1st image in the file.
//P [pgm]	Output in pgm format, note there should only be one input in this case
//P [log]	Logs rotations/translations to align.log
//P [edgenorm]	Perform an edgenormalization of both images before alignment
//P [keepeulers]	Indicates that the input images have accurate Eulers which need to be preserved.
//U align2d proj.hed cls000.hed tst.hed
//D Pretty self explanatory. If the input is a stack, the first image is used.


#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

float angdiff(float alt1,float az1,float alt2,float az2) {
Euler e1(alt1,az1,0);
Euler e2(alt2,az2,0);
return e1.diff(&e2);
}

main(int argc, char *argv[]) {
EMData *r,d,*da;
int i,j,n,pgm=0,noflip=0,edgenorm=0,log=0,fsc=0,noali=0,precen=0,fixn=0,seqali=0,eulali=0,keepeul=0;
float eul[3];

appinit(argc,argv);

if (argc<4) appusage(argv[0]);

for (i=4; i<argc; i++) {	
	if (strncmp(argv[i],"noflip",6)==0) noflip=1;
	else if (strncmp(argv[i],"pgm",3)==0) pgm=1;
	else if (strncmp(argv[i],"log",3)==0) log=1;
	else if (strncmp(argv[i],"fsc",3)==0) fsc=1;
	else if (strncmp(argv[i],"fixn=",5)==0) fixn=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"noali",3)==0) noali=1;
	else if (strncmp(argv[i],"seqali",6)==0) seqali=1;
	else if (strncmp(argv[i],"eulali",6)==0) eulali=1;
	else if (strncmp(argv[i],"precen",6)==0) precen=1;
	else if (strncmp(argv[i],"edgenorm",8)==0) edgenorm=1;
	else if (strncmp(argv[i],"keepeulers",10)==0) keepeul=1;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if (eulali && !precen) error(ERR_EXIT,"eulali works only with the precen option\n");

n=fileCount(argv[2]);
r=new EMData;
r->readImage(argv[1],fixn);
if (edgenorm) r->edgeNormalize();

float *eula = (float *)malloc(n*3*sizeof(float));
// read the Eulers of all the images
if (eulali) {
	for (i=0; i<n; i++) {
		EMData tm;
		tm.readImage(argv[2],i,1);
		eula[i*3]=tm.alt();
		eula[i*3+1]=tm.az();
		eula[i*3+2]=tm.phi();
	}
}

for (i=0; i<n; i++) {
	d.readImage(argv[2],i);
	if (edgenorm) d.edgeNormalize();
	printf("%d. orig %5.2f, %5.2f, %5.2f\t",i,d.alt(),d.az(),d.phi());
	eul[0]=d.alt();
	eul[1]=d.az();
	eul[2]=d.phi();
	if (eulali) {
		float ba1=PI,ba2=PI;
		int bn1=0,bn2=0;
		for (j=0; j<i-1; j++) {
			float ad=angdiff(eula[i*3],eula[i*3+1],eula[j*3],eula[j*3+1]); 
			if (ad<ba1) {
				ba2=ba1;
				bn2=bn1;
				ba1=ad;
				bn1=j;
			}
		}
		da=d.copy();
		if (i && eula[(i-1)*3]==eula[i*3]) bn2=i-1;
		r->readImage(argv[2],bn1);
		da->rotAlignPrecen(r); 
		float aa1=da->alt();
		r->readImage(argv[2],bn2);
		da->rotAlignPrecen(r); 
		if (fabs(aa1-da->alt())>PI) {
			if (aa1>da->alt()) aa1-=2.0*PI;
			else aa1+=2.0*PI;
		}
printf("%d. to \t%d\t%d\t%f\t%f\n",i,bn1,bn2,aa1*180./PI,da->alt()*180./PI);
		aa1=(aa1+da->alt())/2.0;
		da->setRAlign(aa1,0,0);
		da->rotateAndTranslate(); 
				
	}
	else if (noali) da=d.copy(0,0);
	else if (precen) { 
		da=d.copy(); 
		da->rotAlignPrecen(r); 
		da->rotateAndTranslate(); 
	}	
	else if (noflip) da=d.RTAlign(r);
	else {
		da=d.RTFAlign(r);
		da->refineAlign(r);
	}
	printf("dx,dy,da = %1.0f, %1.0f\t%5.2f\n",da->Dx(),da->Dy(),da->alt()*180./PI);
	if (log) {
		FILE *out=fopen("align.log","a");
		if (out) {
			fprintf(out,"%s\t%d\t%d\t%1.2f\t%1.2f\n",argv[2],(int)round(da->Dx()),(int)round(da->Dy()),da->alt()*180/PI,da->ufloat4);
			fclose(out);
		}
	}

	if (fsc) {
		char s[80];
		sprintf(s,"fsc2d.%d.txt",i);
		float *fsc = da->fsc(r);
		save_data(0,1,fsc,da->ySize()/2,s);
		free(fsc);
	}
	if (keepeul) da->setRAlign(eul[0],eul[1],eul[2]+da->alt());
	if (pgm) da->writePGM(argv[3],da->Min(),da->Max());
	else da->writeIMAGIC(argv[3],-1);
	if (seqali) { delete r; r=da; }
	else delete da;
} 

}
