// classesbybasis.C 5/2002  Steve Ludtke

//N classesbybasis
//F This takes an ortogonal basis set as created by svdcmp and uses it to operate on a set of images
//T LM
//1
//P <input>	Input images to classify
//P <ref set>	Reference data set 
//P <basis set>	A set of orthogonal basis images as generated by svdcmp
//P [mask=<rad>]	Mask radius
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [nbasis=<n>]	Number of basis images to use from the basis set. Defaults to all.
//P [logit=<label>]	This specifies a label to use in the particle.log file. This records particle number classification throughout a refinement.
//P [groups=<ngrp>]	This will classify into multiple sets. Used by svdmultirefine.py
//P [maxshift=<dxy>]	Maximum radial translation
//P [verbose]	More verbose output
//P [quiet]	No logging
//U 

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "XYData.h"

struct BEST { float pwr,x,y,ang; int flip; };

#define BESTKP	20		// length of ordered list

// a little class for maintaining an ordered list
class Olist {
public:
BEST *best;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	best=(BEST *)calloc(N,sizeof(BEST));
	for (int i=0; i<N; i++) {
		best[i].pwr=-1.0e23;
	}
}

~Olist() {
	free(best);
}

void insert(float pwr,float x,float y,float ang,int flip) {
	if (pwr<best[N-1].pwr) return;

	int i,j;

	for (i=0; i<N; i++) if (pwr>best[i].pwr) break;

	for (j=N-1; j>i; j--) best[j]=best[j-1];
	
	best[i].pwr=pwr;
	best[i].x=x;
	best[i].y=y;
	best[i].ang=ang;
	best[i].flip=flip;
}

};

inline int negmod(int x,int y) {
if (x>=0) return x%y;
return y+x%y;
}

// This will calculate a set of CCFs for a rotated image, 1 for each basis vector, then
// compute the dot product of the resultant projection vector at each pixel location with
// a set of reference vectors, finally storing the strength of the best match in each pixel location
// in the returned image
EMData *mapbasis(EMData *orig,int nbasis,list *basis,int nclsref,float **clsvec,int maxshift) {
	int i,j,k,l;
	int xs=orig->xSize();
	if (maxshift>xs/4) maxshift=xs/4;
	EMData *ret=orig->copyHead();
	ret->zero();
	
	EMData **ccf = (EMData **)malloc(nbasis*sizeof(EMData *);
	for (i=0; i<nbasis; i++) ccf[i]=orig->calcCCF((EMData *)basis->objectAt(i));
	for (i=-maxshift; i<=maxshift; i++) {
		for (j=-maxshift; j<=maxshift; j++) {
			float bdot=0;
			int bcmp=0;
			for (k=0; k<nclsref; k++) {
				float dot=0;
				for (l=0; l<nbasis; l++) 
					dot+=clsvec[l][k]*ccf[l]->valueAt(negmod(i,xs),negmod(j,xs));
				if (dot>bdot) { bdot=dot; bcmp=-l; }
			}
			ret->setValueAt(negmod(i,xs),negmod(j,xs),0,bdot);
			ret->setValueAt(i+maxshift*2,j+maxshift*2,0,bcmp);		
				// hack. We store the best matching reference number in the unused 1/2 of the map
		}
	}
	
	for (i=0; i<nbasis; i++) delete ccf[i];
	free(ccf);
	
	return ret;
}


main(int argc, char *argv[]) {
int i,j,k,num=0,denom=1,quiet=0,maxshift=-1,precen=0;
int alignn=1,origrt=0,nbasis=-1,maskr=0,verbose=0,groups=0;
float dang=PI;
char *saveali="aligned.hed",*logit=NULL;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=4; i<argc; i++) {	
	if (strncmp(argv[i],"multiali=",9)==0) alignn=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"nbasis=",7)==0) nbasis=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"maxshift=",9)==0) maxshift=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"saveali=",8)==0) saveali=argv[i]+8;
	else if (strncmp(argv[i],"logit=",6)==0) logit=&argv[i][6];
	else if (strncmp(argv[i],"mask=",5)==0) maskr=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else if (strncmp(argv[i],"groups=",7)==0) groups=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (argv[i][7]!='=') verbose=1;
		else verbose=atoi(&argv[i][8]);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

int bc=fileCount(argv[3]);
if (bc<2) error(ERR_EXIT,"Need at least 2 basis images!");
if (nbasis<0) nbasis=bc-1;
if (nbasis>bc) { error(ERR_WARN,"nbasis > # images in basis file"); nbasis=bc-1; }
if (maxshift<=0) precen=1;

// Normal EMAN logging
int Ref;
if (!quiet) {
	LOGbegin(argc,argv);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
	LOG(Ref,argv[2],LOG_REFFILE,NULL);
}

// read basis set and normalize
List *basis=readImages(argv[2],0,nbasis-1);
for (i=0; i<nbasis; i++) {
	EMData *tmp=((EMData *)basis->objectAt(i));
//	tmp->edgeNormalize();
	tmp->multConst(1.0/tmp->SumSq());	// a 'true' normalization
} 
if (maxshift<0) maxshift=((EMData *)basis->objectAt(0))->ySize()/8.0;
dang=atan2(1.0,((EMData *)basis->objectAt(0))->ySize()/2.0);

// read classification refs if present, and determine basis projections
int nclsref=fileCount(argv[2]); 
if (nclsref<2) error(ERR_EXIT,"Could not read 2 or more references from classify file");

// clsvec is a set of nclsref vectors representing projections of the reference set into basis space
float **clsvec;
clsvec=(float **)malloc(nbasis*sizeof(float*));
for (i=0; i<nbasis; i++) clsvec[i]=(float *)malloc(nclsref*sizeof(float)); 

if (clsref) {
	if (verbose) printf("Read classification refs (%s)\n",clsref);

	float *clsmean = (float *)malloc(nbasis*sizeof(float));
	for (j=0; j<nbasis; j++) clsmean[j]=clssig[j]=0;

	// outer loop over classification refs
	for (i=0; i<nclsref; i++) {
		// read reference
		EMData in;
		in.readImage(argv[2],i);
		in.multConst(1.0/in.SumSq());	// normalize this image too
		// determine basis projection for each ref
		for (j=0; j<nbasis; j++) {
			clsvec[j][i]=in.dot((EMData *)basis->objectAt(j));
		}

		if (verbose>1) {
			printf("%d.\t",i);
			for (j=0; j<nbasis; j++) printf("%1.4f\t",clsvec[j][i]);
			printf("\n");
		}
	}
}

if (verbose) printf("Begin analysis:\n");
// Image filtration
// outer loop over images being filtered
int nimg=fileCount(argv[1]);
float *fpdat = NULL;
if (footprint) fpdat=(float *)malloc(nimg*nbasis*sizeof(float));
int start=nimg*num/denom;
int stop=nimg*(num+1)/denom;
for (i=start; i<stop; i++) {
	float dts=0;

	float *vec = (float*)malloc(nbasis*sizeof(float));

	if (verbose>1) printf("\n");

	EMData in;
	in.readImage(argv[1],i);
	in.edgeNormalize();
	Euler eul(*in.getEuler());

	// rot/trans alignment my maximizing basis projection power
	EMData *cop=in.copy();
	float x,y,bx2,by2;
	int nx=in.xSize(),ny=in.ySize();
	int fl,bfl=0;
	float ang=0,bang2;
	BEST *best = (BEST *)calloc(nbasis,sizeof(BEST));
	Olist ol(BESTKP);
	if (!precen) {
		for (fl=0; fl<2; fl++) {
			for (ang=0; ang<PI*2; ang+=dang*6.0) {	//4
				cop->setRAlign(ang);
				cop->setTAlign(0,0);
				cop->rotateAndTranslate();
				if (fl) cop->vFlip();
				EMData *ccfs = mapbasis(cop,nbasis,basis,nclsref,clsvec,maxshift);
				ol.insert(ccfs->Max(),ccfs->MaxLoc()%nx,ccfs->MaxLoc()/nx,ang,fl);
				if (verbose>1) printf("%f\t%f\n",ang*180./PI,ccfs->Max()*1000000000.0);
				delete ccfs;
			}
		}
					
			// this is an ordered list of the best angles we found
			for (j=0; j<BESTKP; j++) {
				if (ol.best[j].x>nx/2) ol.best[j].x=nx-ol.best[j].x;
				else ol.best[j].x*=-1;
				if (ol.best[j].y>ny/2) ol.best[j].y=ny-ol.best[j].y;
				else ol.best[j].y*=-1;
				if (ol.best[j].flip) ol.best[j].y*=-1;
				x= ol.best[j].x*cos(ol.best[j].ang)+ol.best[j].y*sin(ol.best[j].ang);
				y=-ol.best[j].x*sin(ol.best[j].ang)+ol.best[j].y*cos(ol.best[j].ang);
				ol.best[j].x=x;
				ol.best[j].y=y;
			}
		
			if (verbose>2) {
				for (j=0; j<nbasis; j++) 
					printf("%d.\t%1.2f\t%1.2f\t%1.1f\t%c\n",j,best[j].x,best[j].y,best[j].ang*180.0/PI,best[j].flip?'F':' ');
			}
		}
		else {
			x=y=0;
			for (ang=0; ang<2.0*PI; ang+=dang*2.) {
				cop->setRAlign(ang);
				cop->setTAlign(0,0,0);
				cop->rotateAndTranslate();
				float pwr=0;
				for (j=0; j<nbasis; j++) {
					pwr+=SQR(cop->dot((EMData *)basis->objectAt(j)));
					if (pwr>best[j].pwr) { 
						best[j].pwr=pwr; 
						best[j].x=x; 
						best[j].y=y;
						best[j].ang=ang;
						best[j].flip=bfl;
					}
				}
//				if (pwr>bpwr) { bpwr=pwr; bang=ang; bx=x; by=y; }
			}		
		}
		
		// now refine rotation
		k=0;
		while (k<BESTKP && ol.best[k].pwr>ol.best[0].pwr*.5) {
			x=ol.best[k].x;
			y=ol.best[k].y;
			bang2=ol.best[k].ang;
			bfl=ol.best[k].flip;
			for (ang=bang2-dang*4.0; ang<bang2+dang*4.0; ang+=dang/2.) {
				cop->setRAlign(ang);
				cop->setTAlign(x,y,0);
				cop->rotateAndTranslate();
				if (bfl) cop->vFlip();
				float pwr=0;
				for (j=0; j<nbasis; j++) {
					pwr+=SQR(cop->dot((EMData *)basis->objectAt(j)));
					if (pwr>best[j].pwr) { 
						best[j].pwr=pwr; 
						best[j].x=x; 
						best[j].y=y;
						best[j].ang=ang;
						best[j].flip=bfl;
					}
				}
				if (verbose>1) printf("%f\t%f\n",ang*180./PI,pwr);
//			if (pwr>bpwr) { bpwr=pwr; bang=ang; bx=x; by=y; }
			}
			k++;
		}

		if (!precen) {
			for (j=0; j<nbasis; j++) best[j].pwr=0;
			bx2=best[nbasis-1].x;
			by2=best[nbasis-1].y;
			bang2=best[nbasis-1].ang;
			// now refine the dx/dy and rotation
			for (x=bx2-.75; x<=bx2+.76; x+=.25) {
				for (y=by2-.75; y<=by2+.76; y+=.25) {
					for (ang=bang2-dang; ang<bang2+dang; ang+=dang/4.) {
						cop->setRAlign(ang);
						cop->setTAlign(x,y,0);
						cop->rotateAndTranslate();
						if (bfl) cop->vFlip();
						float pwr=0;
						for (j=0; j<nbasis; j++) {
							pwr+=SQR(cop->dot((EMData *)basis->objectAt(j)));
							if (pwr>best[j].pwr) { 
								best[j].pwr=pwr; 
								best[j].x=x; 	
								best[j].y=y;
								best[j].ang=ang;
								best[j].flip=bfl;
							}
						}
//						if (pwr>bpwr) { bpwr=pwr; bang=ang; bx=x; by=y; }
					}
				}
			}
		}
		delete cop;
		if (verbose>1) printf("%d.\t%0.2f\t%0.2f\t%1.1f\t%c\n",i,best[nbasis-1].x,best[nbasis-1].y,best[nbasis-1].ang*180./PI,best[nbasis-1].flip?'F':' ');
		in.setRAlign(best[nbasis-1].ang);
		in.setTAlign(best[nbasis-1].x,best[nbasis-1].y,0);
		in.rotateAndTranslate();
		if (best[nbasis-1].flip) in.vFlip();
		in.setRAlign(eul.alt(),eul.az(),eul.phi()-best[nbasis-1].ang);
		if (maskr) {
			in.maskNormalize(masknorm);
			in.mult(maskc);
		}
	}

	if (saveali) in.writeIMAGIC(saveali,i);

	// inner loop dot product over basis vectors
	for (j=0; j<nbasis; j++) vec[j]=in.dot((EMData *)basis->objectAt(j));

	// This does the classification
	if (clsref) {
		float max=-1.0e30;
		int maxj=0;
		// classify based on basis projection comparisons
		float *ddt=(float *)malloc(nclsref*sizeof(float));
		
		// similarity dot product for each possibility
		for (j=0; j<nclsref; j++) {
			ddt[j]=0;
			if (scalebasis) {
				for (int k=0; k<nbasis; k++) ddt[j]+=vec[k]*clsvec[k][j]*clssig[k];
			} else {
				for (int k=0; k<nbasis; k++) ddt[j]+=vec[k]*clsvec[k][j];
			}
			if (ddt[j]>max) { max=ddt[j]; maxj=j; }
		}

		if (groups) {
			int gsize=1+(bc-1)/groups;	// this should be exactly an integer anyway
			int grp=maxj/gsize;
			char s[200],t[200],u[512];
			sprintf(s,"%0d/cls%04d.lst",grp,maxj%gsize);
			sprintf(t,"%f",max);
			if (saveali) {
				sprintf(u,"../%s",saveali);
				in.writeLST(s,-1,u,i,t);
			}
			else in.writeLST(s,-1,argv[1],i,t);		
		}
		else {
			char s[200],t[200];
			sprintf(s,"cls%04d.lst",maxj);
			sprintf(t,"%f",max);
			if (saveali) in.writeLST(s,-1,saveali,i,t);
			else in.writeLST(s,-1,argv[1],i,t);
		}
			
		if (logit) {
			// order: program tag, particle number, file, ntk rank, quality, defocus, class, iteration tag 
			char s[120];
			float *ctf=in.getCTF();
			if (!in.hasCTF()) ctf[0]=0;
			if (groups) {
				int gsize=bc/groups;	// this should be exactly an integer anyway
				int grp=maxj/gsize;
				char fsp[180];
				char fsp2[1024];
				sprintf(fsp,"%0d/particle.log",grp);
				sprintf(fsp2,"../%s",saveali?saveali:argv[1]);
				sprintf(s,"B\t%d\t%s\t0\t%1.5g\t%1.3f\t%d\t%s\n",i,fsp2,max,-ctf[0],maxj%gsize,logit);
				FSsend(FS_TEXT,strlen(s)+1,fsp,s);
			}
			else {
				sprintf(s,"B\t%d\t%s\t0\t%1.5g\t%1.3f\t%d\t%s\n",i,saveali?saveali:argv[1],max,-ctf[0],maxj,logit);
				FSsend(FS_TEXT,strlen(s)+1,"particle.log",s);
			}
		}
		if (verbose==1||verbose==2) printf("%d -> %d  (%1.4g)\n",i,maxj,max);
		if (verbose>2) {
			printf("%d.",i);
			for (j=0; j<nclsref; j++) printf("\t%1.4f",ddt[j]);
			printf(" -> %d\n",maxj);
		}
		free(ddt);
	}

	free(vec);
	
}

if (!quiet) LOG(Ref,0,LOG_END,NULL);

}
