// ctfavg.C Copyright 1999 Steve Ludtke
// This program will average all the particles in a file together
// including CTF correction. No alignment is performed.

//N ctfavg
//F Averages a set of 2D images with ctf correction
//T LT
//1
//P <infile>	Set of images to be averaged
//P <outfile>	Output, 1 image will be appended
//P [res=<in A>]	Resolution in A if structure factor is not used
//P [sffile=<file>]	Structure factor file used to fit data in infile
//P [mask=<rad>]	Circular mask
//P [copyeuler]	Used only if the filspec is cls####, will read the Eulers from proj.hed
//P [refmask]	Only for cls#### files, will mask based on proj.hed
//P [ref0]	Will ignore the first image in the file
//P [copyrefs]	Copy proj.hed image into the output file
//P [even|odd]	Uses only even or odd numbered particles
//P [snrmult=<val>]	Multiply the SNR by this number when Wiener filtering
//P [nolog]	Do not log a record of this program in .emanlog
//U ctfavg cls0000.lst classes.hed sffile=groel.xray
//D This program will average a set of prealigned images with CTF correction.

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "List.h"

#include "EMData.h"

main(int argc, char** argv)
{
int i,j;
float filtr=25.0;
EMData av;
List *in;
XYData SF;
char *ctfcw=NULL;
int copyeuler=0,refmask=0,mask=0,copyrefs=0,ref0=0,evenodd=0,nolog=0;
float snrmult=1.0;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"res=",4)==0) filtr=atof(&argv[i][4]);
	else if (strncmp(argv[i],"copyeuler",9)==0) copyeuler=1;
	else if (strncmp(argv[i],"copyrefs",8)==0) copyrefs=1;
	else if (strncmp(argv[i],"refmask",7)==0) refmask=1;
	else if (strncmp(argv[i],"usefs",5)==0) ;
	else if (strncmp(argv[i],"quiet",5)==0) nolog=1;
	else if (strncmp(argv[i],"nolog",5)==0) nolog=1;
	else if (strncmp(argv[i],"even",4)==0) evenodd=1;
	else if (strncmp(argv[i],"odd",3)==0) evenodd=2;
	else if (strncmp(argv[i],"ref0",4)==0) ref0=1;
	else if (strncmp(argv[i],"snrmult=",8)==0) snrmult=atof(&argv[i][8]);
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"sffile=",7)==0) {
		ctfcw=&argv[i][7]; 
	}
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

int Ref;
if (!nolog) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

if (ctfcw) {
	SF.readFile(ctfcw);	
	float *d=SF.getData();
	for (i=1; i<SF.n()*2; i+=2) d[i]=(d[i]<=0)?-4.0:log10(d[i]);
	SF.update();


	for (i=0; i<SF.n();i++) 
		if (i==0) printf("x=%f y=%f \n", SF.x(i),SF.y(i));
}

if (evenodd) {
	int ni=fileCount(argv[1]);
	in=new List;
	for (i=ref0+evenodd-1; i<ni; i+=2) {
		EMData *im = new EMData;
		im->readImage(argv[1],i);
		in->addObject(im);
	}
}
else in=readImages(argv[1],ref0,-1);

if (!((EMData *)in->objectAt(0))->hasCTF()) { 
	printf("These images don't appear to have CTF parameters\n"); 
	exit(1); 
}
printf("Averaging %d images\n",in->count());
int nx=((EMData *)in->objectAt(0))->xSize();

av.setNImg(in->count());

EMData tmp;
EMData *masknorm=NULL;
int nf=0;
if (copyeuler || refmask || copyrefs) {
	sscanf(argv[1],"cls%d",&nf);
	tmp.readImage("proj.hed",nf);
	tmp.setNImg(-1);
	if (copyrefs) tmp.writeIMAGIC(argv[2],nf*2);
	if (copyeuler) av.setRAlign(tmp.alt(),tmp.az(),tmp.phi());
	if (refmask) {
	        tmp.realFilter(6);   // value != 0 -> 1
	        masknorm=tmp.copy(0,0);
	        masknorm->addMaskShell(1);
	        masknorm->subtract(&tmp);
	        masknorm->addMaskShell(2);
		for (i=0; i<in->count(); i++) {
			EMData *t=(EMData *)in->objectAt(i);
			t->maskNormalize(masknorm,0);
			t->mult(&tmp);
		}
	}
}
if (mask) {
	tmp.setSize(nx,nx,1);
	tmp.one();
	tmp.applyMask(mask,4);
        masknorm=tmp.copy(0,0);
        masknorm->addMaskShell(1);
        masknorm->subtract(&tmp);
        masknorm->addMaskShell(2);
	for (i=0; i<in->count(); i++) {
		EMData *t=(EMData *)in->objectAt(i);
		t->maskNormalize(masknorm,0);
		t->mult(&tmp);
	}
}

if (snrmult!=1.0) {
	for (i=0; i<in->count(); i++) {
		EMData *t=(EMData *)in->objectAt(i);
		if (!t) error(ERR_EXIT,"No object in SNRMULT");
		float *ctf=t->getCTF();
		if (!ctf) error(ERR_EXIT,"No CTF parms");
		ctf[2]*=snrmult;	// modify the array directly, a bit risky if the library is revamped
	}
}

if (ctfcw) av.makeAverageCTFCW(in,&SF);
else if (filtr>0) av.makeAverageCTFC(in,filtr);
else av.makeAverage(in);
if (refmask ||mask) {
	av.maskNormalize(masknorm,0);
	av.mult(&tmp);
}

if (copyeuler&&copyrefs) av.writeIMAGIC(argv[2],nf*2+1);
else if (copyeuler) av.writeIMAGIC(argv[2],nf);
else av.writeIMAGIC(argv[2],-1);

if (!nolog) {
	LOG(Ref,argv[2],LOG_OUTFILE,NULL);
	LOG(Ref,NULL,LOG_END,NULL);
}
}

