// pdb2mrc.C  11/99  Steve Ludtke
// add {full] [sym=<cn|dn|icos>] [rot=<alt,az,phi>] options on July 2000 by Wen Jiang
// Copyright 1999 Baylor College of Medicine
//
//N pdb2mrc
//F This program will generate a 3D electron density map from PDB files
//T LM
//1
//P <pdb input>	The PDB file to convert
//P <mrc output>	The output file for the generated map
//P [info]	If this is specified, information on the PDB file is displayed, no conversion is performed.
//P [apix=<a/voxel>]	Angstroms per voxel in the output, default is 1
//P [res=<resolution>]	Resolution defined in 'EM' way. reciprocal of the 1/2 width of a Gaussian in Fourier space
//P [box=<box size>]	Box size in voxels, automatic if unspecified
//P [het]	Include HET atoms in the conversion
//P [center]	Center the model (center of charge)
//P [solvate]	Solvate the model (with water). This is an approximate algorithm.
//P	[full]	Apply non-crystallographic symmetry (MTRIXs or BIOMTs) to get full structure
//P	[sym=<cn|dn|icos>]	Apply additional symmetry to the structure
//P	[rot=<alt,az,phi>]	Apply rotation after symmetries are applied,
//P [gen1]	This will generate density for only 1 copy of the molecule, but otherwise act as if all copies were generated.
//U pdb2mrc 1FAQ.pdb FAQ.mrc apix=1 res=1.5 center
//D <b>WARNING: the definition of resolution changed between build 34 and 35</b>
//D This will generate a 3D MRC map from a PDB file. This may be used by <a href=align3d.html>
//E align3d</a> or <a href=foldhunter.html>foldhunter</a>, or for other purposes. The solvate
//E option is better than that used by some other similar programs, but is NOT discrete
//E water molecule solvation. It is still an approximate continuous method which still
//E produces artifacts. In addition, without solvation, the program can generate models at any
//E resolution, but with solvation, resolution must be 1.5 A or better. 
//D	The purpose of [sym=<cn|dn|icos>] is mainly for generating full icosahedral virus structure,
//E	since the PDB files for some viruses (bluetongue virus) only have non-crystallographic symmetry
//E	transform matrix for half of the virus, we need to apply additional symmetry operation (for bluetongue
//E	virus use sym=c2) to generate the full structure. For bluetongue virus, the C2 is 
//E	not enough, it requires another rotation along Z axis to make the the XYZ all on 2-fold
//E	axes, so [rot=<alt,az,phi>] is added for this purpose to tranform the coordinates instead 
//E	tranform of density map later on.

#include "EMData.h"
#include "Vect.h"
#include <vector>
using namespace std;

#define H2ORAD	1.92		// in A
#define H2ODEN	.334		// in e-/A^3

struct ATOM { 
float x,y,z; int a; 
ATOM(float tx, float ty, float tz, float ta){x=tx;y=ty;z=tz;a=ta;};
};

struct MTRIX {
float matrix[3][3];
float trans[3];
MTRIX(){
	matrix[0][0]=1;	matrix[0][1]=0;	matrix[0][2]=0;	
	matrix[1][0]=0;	matrix[1][1]=1;	matrix[1][2]=0;	
	matrix[2][0]=0;	matrix[2][1]=0;	matrix[2][2]=1;	
	trans[0]=0;		trans[1]=0;		trans[2]=0;
}	
bool isIdentity(){
	return(
	matrix[0][0]==1	&&	matrix[0][1]==0	&&	matrix[0][2]==0	&&	
	matrix[1][0]==0	&&	matrix[1][1]==1	&&	matrix[1][2]==0	&&
	matrix[2][0]==0	&&	matrix[2][1]==0	&&	matrix[2][2]==1	&&
	trans[0]==0	&&	trans[1]==0	&&	trans[2]==0	);
}
};	

void main(int argc, char *argv[]) {
float apix=1.0,res=1.5;
int i,j,k,l,box=0,info=0,ires=0,ihelix=0,isheet=0,lres=0,lsheet=0,lhelix=0,het=0;
vector< ATOM > atoms;
vector< MTRIX > transform;
int natoms=0,natoms1=0,naa=32,solv=0,pipe=0,center=0,gen1=0;
char s[1024],lch=' ';
float totmass=0,totchg=0;
float xr[3] = {0,0,0} ,yr[3] = {0,0,0} ,zr[3] = {0,0,0},rmax=0;
bool full=false;
char *sym="c1";
Euler rot(0,0,0);	// default to no rotation as a whole

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {
	if (strncmp(argv[i],"apix=",5)==0) apix=atof(&argv[i][5]);
	else if (strncmp(argv[i],"info",4)==0) info=1;
	else if (strncmp(argv[i],"res=",4)==0) res=atof(&argv[i][4]);
	else if (strncmp(argv[i],"het",3)==0) het=1;
	else if (strncmp(argv[i],"gen1",4)==0) gen1=1;
	else if (strncmp(argv[i],"box=",4)==0) box=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"solvate",7)==0) solv=1;
	else if (strncmp(argv[i],"center",6)==0) center=1;
	else if (strncmp(argv[i],"full",4)==0) full=true;
	else if (strncmp(argv[i],"sym=",4)==0) sym=&argv[i][4];
	else if (strncmp(argv[i],"rot=",4)==0) {
		float alt, az, phi;
		sscanf(&argv[i][4],"%f,%f,%f",&alt,&az,&phi);
		rot.setAngle(alt*PI/180.,az*PI/180.,phi*PI/180.);
	}
	else { fprintf(stderr,"Unknown option %s\n",argv[i]); exit(1); }
}

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

//if (res>3.0) { printf("res > 3.0 !  For low resolutions, use proc3d to apply a low-pass filter\n"); exit(1); }
//if (apix>3.0) { printf("apix > 3.0 !   For less sampling, use proc3d with the shrink option\n"); exit(1); }
if (apix>res) printf("Warning, apix should be <= res\n");
if (solv && (apix>1.0 || res>2.5)) { printf("apix<=1.0 and res<=2.5 for solvation to work properly\n"); exit(1); }

FILE *in;
if (strcmp(argv[1]+strlen(argv[1])-4,".bz2")==0) {
	sprintf(s,"bzcat %s",argv[1]);
	in = popen(s,"r");
	pipe=1;
}
else if (strcmp(argv[1]+strlen(argv[1])-3,".gz")==0) {
	sprintf(s,"zcat %s",argv[1]);
	in = popen(s,"r");
	pipe=1;
}
else in = fopen(argv[1],"r");
if (!in) { printf("Cannot open input file.\n"); exit(1); }

bool header_just_passed=true;
Euler symeuler(0,0,0);
symeuler.setSym(sym);

while ((fgets(s,200,in)!=NULL)) {
	if (strncmp(s,"ENDMDL",6)==0) break;
	if (ihelix>=0 && strncmp(s,"HELIX",5)==0) {
		if (atoi(s+21)>lhelix || s[19]!=lch) ihelix+=abs(atoi(s+21)-atoi(s+33))+1;
		else { printf("Overlapping helices, invalid \n"); ihelix=-1; }
		lch=s[19];
		lhelix=atoi(s+33);
	}
	if (isheet>=0 &&strncmp(s,"SHEET",5)==0) {
		isheet+=abs(atoi(s+22)-atoi(s+33))+1;
	}
	// setup the non-crystallographic symmetry tranformations --Wen
	if (full && strncmp(s,"MTRIX",5)==0) {
		int rowid,transformid;
		float rx,ry,rz,t;

		sscanf(s+5,"%d %d %f %f %f %f",&rowid,&transformid,&rx,&ry,&rz,&t);
		if(transformid>transform.size())transform.resize(transformid);
		transform[transformid-1].matrix[rowid-1][0]=rx;
		transform[transformid-1].matrix[rowid-1][1]=ry;
		transform[transformid-1].matrix[rowid-1][2]=rz;
		transform[transformid-1].trans[rowid-1]=t;
	} 
	if (full && strncmp(s,"REMARK 350   BIOMT",18)==0) {
		int rowid,transformid;
		float rx,ry,rz,t;

		sscanf(s+18,"%d %d %f %f %f %f",&rowid,&transformid,&rx,&ry,&rz,&t);
		//printf("%d:\t%10.6f\t%10.6f\t%10.6f\t\t%10.5f\n",transformid,rx,ry,rz,t);
		
		if(transformid>transform.size())transform.resize(transformid);
		transform[transformid-1].matrix[rowid-1][0]=rx;
		transform[transformid-1].matrix[rowid-1][1]=ry;
		transform[transformid-1].matrix[rowid-1][2]=rz;
		transform[transformid-1].trans[rowid-1]=t;
	} 
	
	if (strncmp(s,"ATOM",4)!=0 && (strncmp(s,"HETATM",6)!=0 || !het)) continue;

	if (strncmp(s,"ATOM",4)==0 && atoi(s+22)!=lres) { lres=atoi(s+22); ires++; }

	if (s[13]==' ') s[13]=s[14];
	if (s[13]==' ') s[13]=s[15];
	
	int e=0;
	switch (s[13]) {
	case 'H': e=1.0; totmass+=1.00794; break;
	case 'C': e=6.0; totmass+=12.0107; break;
	case 'A':	// treat 'A'mbiguous atoms as N, not perfect, but good enough
	case 'N': e=7.0; totmass+=14.00674; break;
	case 'O': e=8.0; totmass+=15.9994; break;
	case 'P': e=15.0; totmass+=30.973761; break;
	case 'S': e=16.0; totmass+=32.066; break;
	case 'W': e=18.0; totmass+=1.00794*2+15.9994; break;	// ficticious water 'atom'
	case 'U': 
		if (s[12]=='A') { e=79.0; totmass+=196.96655; }
		break;	// gold
	default: 
		printf("Unknown atom %c\n",s[13]);
		e=0;
	}
	if (e==0) continue;

	float x, y, z, tr;
	sscanf(&s[28]," %f %f %f",&x,&y,&z);
	
	// check for if the identity transform is included, if not add it  --Wen
	if(header_just_passed) {
		header_just_passed=false;	// only need to check once
		
		bool hasidentity=false;
		for(int sym=0;sym<transform.size();sym++){
			if (transform[sym].isIdentity()){
				hasidentity=true;
			}
		}
		if(! hasidentity){
				transform.resize(transform.size()+1);
		}
	}
	
	if(full && transform.size()>0 ) {
		for(int sym=0;sym<transform.size();sym++){	
			Point p=Point(x,y,z).rotate(transform[sym].matrix)+Point(transform[sym].trans[0],transform[sym].trans[1],transform[sym].trans[2]);
			atoms.push_back(ATOM(p.X(),p.Y(),p.Z(),e));
			natoms++;
	  	}
	} else {
			atoms.push_back(ATOM(x,y,z,e));
			natoms++;
	}
}
if (pipe) pclose(in);
else fclose(in);
if (natoms==0 || ires==0) { printf("no data read!\n"); exit(1); }
natoms1=natoms;

// apply additional symmetry, primarily for making full virus (bluetongue virus needs sym=c2)  --Wen
if(symeuler.getMaxSymEl()>1){
	int num_atoms=natoms;
	totmass*=symeuler.getMaxSymEl();
	for(int i=0;i<num_atoms;i++){
		Point p=Point(atoms[i].x,atoms[i].y,atoms[i].z);
		float tr, e=atoms[i].a;
		symeuler.resetSym();symeuler.NextSym();
		for(Euler symel=symeuler.NextSym();	// start from the 2nd, since the 1st is already there 
				  symel.valid(); symel=symeuler.NextSym()) {
			Point tmpp=p.rotate(symel);
			atoms.push_back(ATOM(tmpp.X(),tmpp.Y(),tmpp.Z(),e));
			natoms++;
		}
	}
}

// now apply the rotation on all the atoms if neccesary, (bluetongue virus needs rot=0,88,0)  --Wen
if(rot.alt()!=0 || rot.az()!=0 || rot.phi()!=0){ // only when rot is not identity
	for(int i=0;i<atoms.size();i++){
		Point tmp=Point(atoms[i].x,atoms[i].y,atoms[i].z).rotate(rot);
		atoms[i].x=tmp.X();	atoms[i].y=tmp.Y();	atoms[i].z=tmp.Z();	
	}	
}

// now calculate the stats
xr[0]=atoms[0].x;
xr[1]=atoms[0].x;
yr[0]=atoms[0].y;
yr[1]=atoms[0].y;
zr[0]=atoms[0].z;
zr[1]=atoms[0].z;
if(atoms.size()>1){
	float tr;
	for(int i=0;i<atoms.size();i++){
		if ((tr=hypot3(atoms[i].x,atoms[natoms].y,atoms[natoms].z))>rmax) 
			rmax=tr;

 	 	if (atoms[i].x<xr[0]) xr[0]=atoms[i].x;
 	 	if (atoms[i].x>xr[1]) xr[1]=atoms[i].x;
 	 	if (atoms[i].y<yr[0]) yr[0]=atoms[i].y;
 	 	if (atoms[i].y>yr[1]) yr[1]=atoms[i].y;
 	 	if (atoms[i].z<zr[0]) zr[0]=atoms[i].z;
 	 	if (atoms[i].z>zr[1]) zr[1]=atoms[i].z;
		xr[2]+=atoms[i].x*atoms[i].a;
		yr[2]+=atoms[i].y*atoms[i].a;
		zr[2]+=atoms[i].z*atoms[i].a;
		totchg+=atoms[i].a;
	}
	xr[2]/=totchg;
	yr[2]/=totchg;
	zr[2]/=totchg;
}	

printf("%d atoms read\n",natoms);
printf("Total Mass = %1.1f Da, Total Charge = %1.1f e-\n",totmass,totchg);
printf("Limits (in A) %1.1f, %1.1f, %1.1f - %1.1f, %1.1f, %1.1f\n",xr[0],yr[0],zr[0],xr[1],yr[1],zr[1]);
printf("Center of charge = %1.1f,%1.1f,%1.1f\n",xr[2],yr[2],zr[2]);
if (center) {
	printf("Adjust center -> 0,0,0\n");
	
	rmax=0;
	for (i=0; i<natoms; i++) { 
		atoms[i].x-=xr[2]; 
		atoms[i].y-=yr[2]; 
		atoms[i].z-=zr[2]; 
		if (hypot3(atoms[i].x,atoms[i].y,atoms[i].z)>rmax) 
			rmax=hypot3(atoms[i].x,atoms[i].y,atoms[i].z);
	}
	xr[0]-=xr[2];
	xr[1]-=xr[2];
	yr[0]-=yr[2];
	yr[1]-=yr[2];
	zr[0]-=zr[2];
	zr[1]-=zr[2];
	xr[2]=yr[2]=zr[2]=0;
}
rmax=sqrt(rmax);

char *ss=(char *)malloc(strlen(argv[1])+1);
strcpy(ss,argv[1]);
for (i=0; i<strlen(ss); i++) if (ss[i]=='.') ss[i]=0;

int totres=ires*symeuler.getMaxSymEl();
if(full && transform.size()>1)totres*=transform.size();
printf("%s Max radius = %1.2f\t%d residues\t%1.1f%% sheet\t%1.1f%% helix\n",ss,rmax,totres,isheet*100.0/ires,ihelix*100.0/ires);
if (info) exit(1);

if (box<5) {
	box=ceil(rmax);
	box/=apix;
	box=4*((box-1)/4+2); // round up to nearest 4 then add 4
	box*=2;
}
printf("Box size %d\n",box);

EMData *data,*rdata;
float *d,*rd;

data=new EMData;
data->setSize(box,box,box);
data->zero();

if (solv) {
	rdata=new EMData;
	rdata->setSize(box,box,box);
	rdata->one();
	rdata->multConst(100.0);
	rd=rdata->getData();
}

d=data->getData();

int x[2],y[2],z[2],w;
float xx,yy,zz;
float r,rp,kn;

rp=res/apix;
rp=SQR(PI/rp);		// constant for gaussian falloff
kn=1.0/pow(res*res*PI/2,1.5);	// normalization constant
w=round(res*3.0/apix);
if (w<3) { printf("You do not have sufficient sampling for this resolution. Decrease apix.\n"); exit(1); }
for (l=0; l<(gen1?natoms1:natoms); l++) {
	xx=(atoms[l].x/apix)+box/2;
	yy=(atoms[l].y/apix)+box/2;
	zz=(atoms[l].z/apix)+box/2;
	x[0]=round(xx-w);
	x[1]=round(xx+w);
	y[0]=round(yy-w);
	y[1]=round(yy+w);
	z[0]=round(zz-w);
	z[1]=round(zz+w);
	
	
	if (x[0]>=box||y[0]>=box||z[0]>=box||x[1]<0||y[1]<0||z[1]<0) continue;
	if (x[0]<0) x[0]=0;
	if (x[1]>=box) x[1]=box-1;
	if (y[0]<0) y[0]=0;
	if (y[1]>=box) y[1]=box-1;
	if (z[0]<0) z[0]=0;
	if (z[1]>=box) z[1]=box-1;
	
	for (i=x[0]; i<x[1]; i++) {
		for (j=y[0]; j<y[1]; j++) {
			for (k=z[0]; k<z[1]; k++) {
				r=hypot3((float)i-xx,(float)j-yy,(float)k-zz);
				if (solv) {
					if (rd[i+j*box+k*box*box]>r) rd[i+j*box+k*box*box]=r;
				}
				d[i+j*box+k*box*box]+=kn*atoms[l].a*exp(-r*rp);
			}
		}
	}
}


if (solv) {
	for (i=0; i<box*box*box; i++) if (rd[i]==100.0) rd[i]=-100.0;
	
	// iteratively locate 'water' regions
	int l,c=1,n,i2,j2,k2;
	while (c) {
		c=0;
		for (i=1; i<box-1; i++) {
			for (j=1; j<box-1; j++) {
				for (k=1; k<box-1; k++) {

					// this counts points in a 3x3x3 cube that are already 'water'
					n=0;
					for (i2=i-1; i2<i+2; i2++) {
						for (j2=j-1; j2<j+2; j2++) {
							for (k2=k-1; k2<k+2; k2++) {
								l=i2+j2*box+k2*box*box;
								if (rd[l]<0) n++;
							}
						}
					}
					
					// if 'water' is in at least 6 neighbors and we have a large radius, we're also water
					l=i+j*box+k*box*box;
					if (rd[l]>H2ORAD/apix && d[l]<H2ODEN && n>6) { rd[l]=-rd[l]; c=1; }
				}
			}
		}
	}
	
	for (i=0; i<box*box*box; i++) {
		rd[i]=-rd[i];
//		if (rd[i]>1.92/apix) d[i]=H2ODEN;		// bulk water
//		else if (rd[i]>.7/apix) d[i]+=exp(-(1.92/apix-rd[i])/rp)*kn*18.0;		//more than 1/2 a C bond dist
		if (rd[i]>SQR(2.0*res/apix)) {
			d[i]=H2ODEN;
//			printf("%d\t%f\n",i,rd[i]);
		}
		else if (rd[i]>SQR(H2ORAD/apix)&&d[i]<H2ODEN) {
//			float rr=(sqrt(rd[i])-H2ORAD/apix)/(2.0*res/apix-H2ORAD/apix);
			d[i]=H2ODEN;
//			d[i]+=H2ODEN*exp(-rr*rr);
//printf("%1.1f\n",rr);
//			rr=exp(-3.0*rr*rr);
//			d[i]=rr*H2ODEN+(1.0-rr)*d[i];
		}
	}
	rdata->doneData();
}

data->doneData();
data->setPixel(apix);
data->writeMRC(argv[2]);

LOG(Ref,argv[2],LOG_OUTFILE,NULL);
LOG(Ref,NULL,LOG_END,NULL);
//if (solv) rdata->writeMRC("rad.mrc");
}
