// projtree.C  10/99  Steve Ludtke

//N projtree
//F Generates a list of projections used for high speed particle classification
//T LM
//1
//P <infile>	Projection file
//P <outfile>	Output 'tree' text file
//P [level=2,3]	The decomposition level to use
//P [sym=<cn,icos,etc]	Model symmetry, used for azimuth wrapping
//U projtree proj.img proj.tree level=2 sym=c4
//D This program is used internally by <a href=refine.html>refine</a>. It generates a file used by 
//E <a href=classesbymra.html>classesbymra</a>. Normally classesbymra comapres each particle to
//E each projection. The 'tree' option allows each particle to be compared to 1/4 (level 2) or
//E 1/9 (level 3) of the projections initially. Each particle is then locally refined to
//E determine the exact matching projection. So, the total number of comparisons is N/4 + 6
//E or N/9 + 18. This is an improvement only if the number of projections is sufficiently large.
//E There is also a potential problem if the images are very noisy and the 'best' peaks are
//E sharp and poorly defined. If the incorrect initial maxima is chosen, the 'detailed' maxima
//E will also obviously be wrong. 

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

// a little class for maintaining an ordered list
class Olist {
public:
float *r;
int *n;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	r=(float *)malloc(N*sizeof(float));
	n=(int *)malloc(N*sizeof(int));
	for (int i=0; i<N; i++) {
		r[i]=1.0e20;
		n[i]=-1;
	}
}

~Olist() {
	free(r);
	free(n);
}

void insert(float vr, int vn) {
	int i,j;
	
	for (i=0; i<N; i++) if (r[i]>vr) break;

	if (i==N) return;

	for (j=N-1; j>i; j--) {
		r[j]=r[j-1];
		n[j]=n[j-1];
	}
	r[i]=vr;
	n[i]=vn;
}

};

float csub(float range,float a1,float a2) {
float r;
a1=a1-range*floor(a1/range);
a2=a2-range*floor(a1/range);
r=fabs(a1-a2);
if (r>range/2.0) return range-r;
return r;
}

main(int argc, char *argv[]) {
int i,j,k,csym,level=2;
float az1=2.0*PI;
FILE *out;
EMData *d;
List *all;

appinit(argc,argv);

if (argc<2) appusage(argv[0]);

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"level=",6)==0) level=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"sym=",4)==0) {
		if (tolower(argv[i][4])=='c' || tolower(argv[i][4])=='d' ) {
			csym=atoi(&argv[i][5]);
			az1/=(float)csym;
		}
		else if (argv[i][4]=='h' || argv[i][4]== 'H') {
			i=atof(&argv[i][5]);
			az1=i*PI/180.0;
		}
		else if (strncmp(argv[i]+4,"icos",4)==0) {
			csym=60;
			az1/=5.0;
		}
		else printf("Unsupported symmetry!\n");
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (level<2 || level>3) { printf("Level must be 2 or 3\n"); exit(1); }

out=fopen(argv[2],"w");
if (!out) { printf("Cannot open output file\n"); exit(1); }

all=readImages(argv[1],-1,-1,1);	// headers only

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);
LOG(Ref,argv[2],LOG_OUTFILE,NULL);

int n=all->count();
if (n<7) { printf("Not enough images\n"); exit(1); }

int ur=0;
float lalt=0;
EMData *d1,*d2;
for (i=0; i<n; i++) {
	d1=(EMData *)all->objectAt(i);
	if (!ur && i%level==0) {
		if (level==2) k=6;
		else if (level==3) k=18;

		Olist ol(k);
		for (j=0; j<n; j++) {
			if (j==i) continue;
			d2=(EMData *)all->objectAt(j);
			ol.insert(hypot(d1->alt()-d2->alt(),csub(az1,d1->az(),d2->az())),j);
		}
		fprintf(out,"%d\t",i);
		for (j=0; j<k; j++) fprintf(out,"%d ",ol.n[j]);
		fprintf(out,"\n");
	}

	if (lalt!=d1->alt()) {
		lalt=d1->alt();
		ur++;
		ur=ur%level;
	}
}
fclose(out);
LOG(Ref,NULL,LOG_END,NULL);
}
